import os
import sys

from .cli import main

if __name__ == '__main__':  # pragma: no cover
    try:
        sys.exit(main() or 0)
    except Exception:
        if os.getenv('PPROTECT_ENABLE_DEBUG'):
            import pdb;pdb.post_mortem()
        raise



# -*- coding: utf-8 -*-
"""
pocket_protector

People-centric secret management system, built to work with modern distributed version control systems.
"""
# Note that the doc above is part of "pprotect -h" output, add to it wisely.

import os
import sys
import json
import difflib

from face import Command, Flag, face_middleware, CommandLineError, UsageError, echo, prompt

from ._version import __version__
from .file_keys import KeyFile, Creds, PPError

_ANSI_FORE_RED = '\x1b[31m'
_ANSI_FORE_GREEN = '\x1b[32m'
_ANSI_RESET_ALL = '\x1b[0m'

# TODO: custodian-signed values. allow custodians to sign values
# added/set by others, then produced reports on which secrets have been
# updated/changed but not signed yet. enables a review/audit mechanism.

try:
    unicode
except NameError:
    # py3
    unicode = str


def _get_text(inp):
    if not isinstance(inp, unicode):
        return inp.decode('utf8')
    return inp


def _create_protected(path):
    if os.path.exists(path):
        raise UsageError('Protected file already exists: %s' % path, 2)
    open(path, 'wb').close()
    kf = KeyFile(path=path)
    # TODO: add audit log entry for creation date
    # TODO: add audit log dates in general
    kf.write()
    return kf


def _ensure_protected(path):
    if not os.path.exists(path):
        raise UsageError('Protected file not found: %s' % path, 2)
    kf = KeyFile.from_file(path)
    return kf


def _get_colorized_lines(lines):
    ret = []
    colors = {'-': _ANSI_FORE_RED, '+': _ANSI_FORE_GREEN}
    for line in lines:
        if line[0] in colors:
            line = colors[line[0]] + line + _ANSI_RESET_ALL
        ret.append(line)
    return ret


def _get_new_creds(confirm=True):
    user_id = prompt('User email')
    passphrase = prompt.secret('Passphrase: ', confirm=confirm)
    ret = Creds(user_id, passphrase)
    return ret


def _get_creds(kf,
               user=None,
               interactive=True,
               check_env=True,
               passphrase_file=None,
               user_env_var='PPROTECT_USER',
               pass_env_var='PPROTECT_PASSPHRASE'):
    if not interactive and not check_env:
        raise UsageError('expected at least one of check_env'
                         ' and interactive to be True', 2)
    user_source = 'argument'
    passphrase, passphrase_source = None, None
    if passphrase_file:
        passphrase_file = os.path.abspath(passphrase_file)
        try:
            passphrase = open(passphrase_file, 'rb').read().decode('utf8')
        except IOError as ioe:
            if getattr(ioe, 'strerror', None):
                msg = '%s while reading passphrase from file at "%s"' % (ioe.strerror, passphrase_file)
            else:
                msg = 'Failed to read passphrase from file at "%s"' % passphrase_file
            raise UsageError(msg=msg)
        else:
            passphrase_source = "passphrase file: %s" % passphrase_file
    if user is None and user_env_var:
        user = os.getenv(user_env_var)
        user_source = 'env var: %s' % user_env_var
    if passphrase is None and pass_env_var:
        passphrase = os.getenv(pass_env_var)
        passphrase_source = 'env var: %s' % pass_env_var

    if interactive:
        if user is None:
            user = prompt('User email')
            user_source = 'stdin'
        if passphrase is None:
            passphrase = prompt.secret('Passphrase: ', confirm=False)
            passphrase_source = 'stdin'

    creds = Creds(_get_text(user), _get_text(passphrase),
                  name_source=user_source, passphrase_source=passphrase_source)
    _check_creds(kf, creds)

    return creds


def _check_creds(kf, creds):
    if kf.check_creds(creds):
        return True

    msg = 'Invalid user email'
    if creds.name_source:
        msg += ' (from %s)' % creds.name_source
    msg += ' or passphrase'
    if creds.passphrase_source:
        msg += ' (from %s)' % creds.passphrase_source
    msg += '. Check credentials and try again.'
    empty_fields = []
    if creds.name == '':
        empty_fields.append('user ID')
    if creds.passphrase == '':
        empty_fields.append('passphrase')
    if empty_fields:
        msg += ' (Warning: Empty ' + ' and '.join(empty_fields) + '.)'

    raise UsageError(msg, 1)


def _get_cmd(prepare=False):
    cmd = Command(name='pocket_protector', func=None, doc=__doc__)  # func=None means output help

    # add flags
    cmd.add('--file', missing='protected.yaml',
            doc='path to the PocketProtector-managed file, defaults to protected.yaml in the working directory')
    cmd.add('--confirm', parse_as=True,
            doc='show diff and prompt for confirmation before modifying the file')
    cmd.add('--non-interactive', parse_as=True,
            doc='disable falling back to interactive authentication, useful for automation')
    cmd.add('--ignore-env', parse_as=True, display=False,  # TODO: keep?
            doc='ignore environment variables like PPROTECT_PASSPHRASE')
    cmd.add('--user', char='-u',
            doc="the acting user's email credential")
    cmd.add('--passphrase-file',
            doc='path to a file containing only the passphrase, likely provided by a deployment system')

    # add middlewares, outermost first ("first added, first called")
    cmd.add(mw_verify_creds)
    cmd.add(mw_write_kf)
    cmd.add(mw_ensure_kf)
    cmd.add(mw_exit_handler)

    # add subcommands
    cmd.add(add_key_custodian, name='init', doc='create a new protected')
    cmd.add(add_key_custodian)

    cmd.add(add_domain)
    cmd.add(rm_domain)

    cmd.add(add_owner)
    cmd.add(rm_owner)

    cmd.add(add_secret)
    cmd.add(update_secret)
    cmd.add(rm_secret)

    cmd.add(set_key_custodian_passphrase)
    cmd.add(rotate_domain_keys)

    cmd.add(decrypt_domain, posargs={'count': 1, 'provides': 'domain_name'})

    cmd.add(list_domains)
    cmd.add(list_domain_secrets, posargs={'count': 1, 'provides': 'domain_name'})
    cmd.add(list_all_secrets)
    cmd.add(list_audit_log)

    cmd.add(print_version, name='version')

    if prepare:
        cmd.prepare()  # an optional check on all subcommands, not just the one being executed

    return cmd


def main(argv=None):  # pragma: no cover  (see note in tests.test_cli.test_main)
    cmd = _get_cmd()

    cmd.run(argv=argv)  # exit behavior is handled by mw_exit_handler

    return


"""
The following subcommand handlers all update/write to a protected file (wkf).
"""

def add_key_custodian(wkf):
    'add a new key custodian to the protected'
    echo('Adding new key custodian.')
    creds = _get_new_creds()
    return wkf.add_key_custodian(creds)


def add_domain(wkf, creds):
    'add a new domain to the protected'
    echo('Adding new domain.')
    domain_name = prompt('Domain name')

    return wkf.add_domain(domain_name, creds.name)


def rm_domain(wkf):
    'remove a domain and all of its keys from the protected'
    echo('Removing domain.')
    domain_name = prompt('Domain name')
    return wkf.rm_domain(domain_name)


def add_owner(wkf, creds):
    'add a key custodian to the owner list of a specific domain'
    echo('Adding domain owner.')
    domain_name = prompt('Domain name')
    new_owner_name = prompt('New owner email')
    return wkf.add_owner(domain_name, new_owner_name, creds)


def rm_owner(wkf):
    'remove a key custodian from the owner list of a domain'
    echo('Removing domain owner.')
    domain_name = prompt('Domain name')
    owner_name = prompt('Owner email')
    return wkf.rm_owner(domain_name, owner_name)


def add_secret(wkf):
    'add a secret to a domain'
    echo('Adding secret value.')
    domain_name = prompt('Domain name')
    secret_name = prompt('Secret name')
    secret_value = prompt('Secret value')
    return wkf.add_secret(domain_name, secret_name, secret_value)


def update_secret(wkf):
    'update a secret value in a domain'
    echo('Updating secret value.')
    domain_name = prompt('Domain name')
    secret_name = prompt('Secret name')
    secret_value = prompt('Secret value')
    return wkf.update_secret(domain_name, secret_name, secret_value)


def rm_secret(wkf):
    'remove a secret from a domain'
    echo('Updating secret value.')
    domain_name = prompt('Domain name')
    secret_name = prompt('Secret name')
    return wkf.rm_secret(domain_name, secret_name)


def set_key_custodian_passphrase(wkf):
    'update a key custodian passphrase'
    user_id = prompt('User email')
    passphrase = prompt.secret('Current passphrase: ')
    creds = Creds(user_id, passphrase)
    _check_creds(wkf, creds)
    new_passphrase = prompt.secret('New passphrase: ', confirm=True)
    return wkf.set_key_custodian_passphrase(creds, new_passphrase)


def rotate_domain_keys(wkf, creds):
    'rotate the internal encryption keys for a given domain'
    domain_name = prompt('Domain name')
    return wkf.rotate_domain_key(domain_name, creds)


"""
Read-only operations follow
"""

def print_version():
    'print the PocketProtector version and exit'
    echo('pocket_protector version %s' % __version__)
    sys.exit(0)


def decrypt_domain(kf, creds, domain_name):
    'output the decrypted contents of a domain in JSON format'
    decrypted_dict = kf.decrypt_domain(domain_name, creds)
    echo(json.dumps(decrypted_dict, indent=2, sort_keys=True))
    return 0


def list_domains(kf):
    'print a list of domain names, if any'
    domain_names = kf.get_domain_names()
    if domain_names:
        echo('\n'.join(domain_names))
    else:
        echo.err('(No domains in protected at %s)' % kf.path)
    return


def list_domain_secrets(kf, domain_name):
    'print a list of secret names for a given domain'
    secret_names = kf.get_domain_secret_names(domain_name)
    if secret_names:
        echo('\n'.join(secret_names))
    else:
        echo.err('(No secrets in domain %r of protected at %s)'
                 % (domain_name, kf.path))
    return


def list_all_secrets(kf):
    'print a list of all secret names, along with the domains that define each'
    secrets_map = kf.get_all_secret_names()
    if not secrets_map:
        echo.err('(No secrets in protected at %s)' % kf.path)
    else:
        for secret_name in sorted(secrets_map):
            domain_names = sorted(set(secrets_map[secret_name]))
            echo('%s: %s' % (secret_name, ', '.join(domain_names)))
    return


def list_audit_log(kf):
    'print a list of actions from the audit log, one per line'
    log_list = kf.get_audit_log()
    echo('\n'.join(log_list))
    return


"""
End subcommand handlers

Begin middlewares
"""


@face_middleware(provides=['creds'], optional=True)
def mw_verify_creds(next_, kf, user, ignore_env, non_interactive, passphrase_file):
    creds = _get_creds(kf, user,
                       check_env=not ignore_env,
                       interactive=not non_interactive,
                       passphrase_file=passphrase_file)
    return next_(creds=creds)


@face_middleware(provides=['kf'], optional=True)
def mw_ensure_kf(next_, file, subcmds_):
    file_path = file or 'protected.yaml'
    file_abs_path = os.path.abspath(file_path)
    init_kf = subcmds_[0] == 'init'
    if init_kf:
        kf = _create_protected(file_abs_path)
    else:
        kf = _ensure_protected(file_abs_path)

    try:
        ret = next_(kf=kf)
    except:
        if init_kf:
            try:
                os.unlink(file_abs_path)
            except Exception:
                echo.err('Warning: failed to remove file: %s' % file_abs_path)
        raise

    return ret


@face_middleware(provides=['wkf'], optional=True)
def mw_write_kf(next_, kf, confirm):
    if not os.access(kf.path, os.W_OK):
        raise UsageError('expected %r to be a writable file. Check the'
                         ' permissions and try again.' % kf.path)

    modified_kf = next_(wkf=kf)

    if not modified_kf:
        return modified_kf

    if confirm:
        diff_lines = list(difflib.unified_diff(kf.get_contents().splitlines(),
                                               modified_kf.get_contents().splitlines(),
                                               kf.path + '.old', kf.path + '.new'))
        diff_lines = _get_colorized_lines(diff_lines)
        echo('Changes to be written:\n')
        echo('\n'.join(diff_lines) + '\n')
        do_write = prompt('Write changes? [y/N] ')
        if not do_write.lower().startswith('y'):
            echo('Aborting...')
            sys.exit(0)

    modified_kf.write()

    return


@face_middleware
def mw_exit_handler(next_):
    status = 55  # should always be set to something else
    try:
        try:
            status = next_() or 0
        except PPError as ppe:
            raise UsageError(ppe.args[0])
    except KeyboardInterrupt:
        echo('')
        status = 130
    except EOFError:
        echo('')
        status = 1

    sys.exit(status)

    return




import nacl
import pytest
import pocket_protector.file_keys

@pytest.fixture
def _fast_crypto():
    old_opslimit = pocket_protector.file_keys.OPSLIMIT
    old_memlimit = pocket_protector.file_keys.MEMLIMIT

    pocket_protector.file_keys.OPSLIMIT = nacl.pwhash.OPSLIMIT_MIN
    pocket_protector.file_keys.MEMLIMIT = nacl.pwhash.MEMLIMIT_MIN

    yield

    pocket_protector.file_keys.OPSLIMIT = old_opslimit
    pocket_protector.file_keys.MEMLIMIT = old_memlimit
    return



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
import json
import subprocess

import ruamel.yaml
from face import CommandChecker

from pocket_protector import cli


def test_prepare():
    # confirms that all subcommands compile together nicely
    assert cli._get_cmd(prepare=True)
    return

KURT_EMAIL = 'kurt@example.com'
KURT_PHRASE = u'passphrasë'
MH_EMAIL = 'mahmoud@hatnote.com'
MH_PHRASE = 'thegame'
DOMAIN_NAME = 'first-domain'
SECRET_NAME = 'secret-name'
SECRET_VALUE = u'secrët-value'


# _fast_crypto from conftest
def test_cli(tmp_path, _fast_crypto):
    cmd = cli._get_cmd()
    cc = CommandChecker(cmd, reraise=True)

    assert cc.run('pprotect version').stdout.startswith('pocket_protector version')

    tmp_path = str(tmp_path)
    protected_path = tmp_path + '/protected.yaml'

    # fail init and ensure that file isn't created
    cc.fail_1('pprotect init --file %s' % protected_path,
              input=[KURT_EMAIL, KURT_PHRASE, KURT_PHRASE + 'nope'])
    assert not os.path.exists(protected_path)

    # successfully create protected
    cc.run('pprotect init --file %s' % protected_path,
           input=[KURT_EMAIL, KURT_PHRASE, KURT_PHRASE])

    # check we can only create it once
    res = cc.fail_2('pprotect init --file %s' % protected_path,
                    input=[KURT_EMAIL, KURT_PHRASE, KURT_PHRASE])

    file_data = ruamel.yaml.YAML().load(open(protected_path).read())
    assert list(file_data['key-custodians'])[0] == KURT_EMAIL
    assert len(file_data['audit-log']) == 1

    res = cc.run('pprotect list-audit-log --file %s' % protected_path)
    assert len(res.stdout.splitlines()) == 1

    # make a new cc, with env and tmp_path baked in (also tests
    # protected.yaml in the cur dir being the default file)
    kurt_env = {'PPROTECT_USER': KURT_EMAIL, 'PPROTECT_PASSPHRASE': KURT_PHRASE}
    cc = CommandChecker(cmd, chdir=tmp_path, env=kurt_env, reraise=True)

    cc.run(['pprotect', 'add-domain'], input=[DOMAIN_NAME])
    res = cc.run(['pprotect', 'list_domains'])
    assert res.stdout.splitlines() == [DOMAIN_NAME]

    cc.run(['pprotect', 'add-secret'],
           input=[DOMAIN_NAME, SECRET_NAME, 'tmpval'])
    cc.run(['pprotect', 'update-secret'],
           input=[DOMAIN_NAME, SECRET_NAME, SECRET_VALUE])
    res = cc.run(['pprotect', 'list-domain-secrets', DOMAIN_NAME])
    assert res.stdout == SECRET_NAME + '\n'

    res = cc.run(['pprotect', 'decrypt-domain', DOMAIN_NAME])
    res_data = json.loads(res.stdout)
    assert res_data[SECRET_NAME] == SECRET_VALUE

    cc.fail(['pprotect', 'decrypt-domain', 'nonexistent-domain'])

    # already exists
    cc.fail_1('pprotect add-key-custodian', input=[KURT_EMAIL, ''])

    cc.run('pprotect add-key-custodian', input=[MH_EMAIL, MH_PHRASE, MH_PHRASE])

    cc.run('pprotect add-owner', input=[DOMAIN_NAME, MH_EMAIL])

    # missing protected
    cc.fail_2('pprotect list-all-secrets', chdir=tmp_path + '/..')

    cc.run('pprotect list-all-secrets')
    assert SECRET_NAME in res.stdout

    cc.run(['pprotect', 'rotate_domain_keys'], input=[DOMAIN_NAME])


    # test mixed env var and entry
    cc.run(['pprotect', 'decrypt-domain', DOMAIN_NAME],
           env={'PPROTECT_USER': MH_EMAIL, 'PPROTECT_PASSPHRASE': None},
           input=[MH_PHRASE])
    assert json.loads(res.stdout)[SECRET_NAME] == SECRET_VALUE

    # test bad creds
    cc.fail_1(['pprotect', 'decrypt-domain', DOMAIN_NAME],
              env={'PPROTECT_USER': None, 'PPROTECT_PASSPHRASE': 'nope'},
              input=[KURT_EMAIL])

    res = cc.fail_1('pprotect set-key-custodian-passphrase',
                    input=[KURT_EMAIL, KURT_PHRASE, KURT_PHRASE, KURT_PHRASE + 'nope'])
    assert 'did not match' in res.stderr

    # correctly reset passphrase
    new_kurt_phrase = KURT_PHRASE + 'yep'
    res = cc.run('pprotect set-key-custodian-passphrase',
                 input=[KURT_EMAIL, KURT_PHRASE, new_kurt_phrase, new_kurt_phrase])

    # try new passphrase with a passphrase file why not
    ppfile_path = str(tmp_path) + 'tmp_passphrase'
    with open(ppfile_path, 'wb') as f:
        f.write(new_kurt_phrase.encode('utf8'))
    res = cc.run(['pprotect', 'decrypt-domain', '--non-interactive',
                  '--passphrase-file', ppfile_path, DOMAIN_NAME])

    res_data = json.loads(res.stdout)
    assert res_data[SECRET_NAME] == SECRET_VALUE

    # test mutual exclusivity of check env and interactive
    cc.fail_2(['pprotect', 'decrypt-domain',
               '--non-interactive', '--ignore-env', DOMAIN_NAME])

    # test removals
    cc.run(['pprotect', 'rm-owner'], input=[DOMAIN_NAME, MH_EMAIL])
    cc.run(['pprotect', 'rm-secret'], input=[DOMAIN_NAME, SECRET_NAME])
    cc.run(['pprotect', 'rm-domain', '--confirm'], input=[DOMAIN_NAME, 'y'])


def test_main(tmp_path):
    # TODO: pytest-cov knows how to make coverage work across
    # subprocess boundaries...
    os.chdir(str(tmp_path))
    res = subprocess.check_output(['pprotect', 'version'])
    assert res.decode('utf8').startswith('pocket_protector version')

    res = subprocess.check_output(['pocket_protector', 'version'])
    assert res.decode('utf8').startswith('pocket_protector version')




import os
import imp
from setuptools import setup, find_packages

__author__ = "Kurt Rose and Mahmoud Hashemi"
__contact__ = "kurt@kurtrose.com"
__license__ = 'Apache License 2.0'
__url__ = 'https://github.com/SimpleLegal/pocket_protector'

CUR_PATH = os.path.abspath(os.path.dirname(__file__))
_version_mod_path = os.path.join(CUR_PATH, 'pocket_protector', '_version.py')
_version_mod = imp.load_source('_version', _version_mod_path)
__version__ = _version_mod.__version__

with open('README.md') as readme:
    long_description = readme.read()


setup(
    name="pocket-protector",
    description="Handy secret management system with a convenient CLI and readable storage format.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=__author__,
    author_email=__contact__,
    url=__url__,
    license=__license__,
    platforms='any',
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={'console_scripts': ['pprotect = pocket_protector.__main__:main',
                                      'pocket_protector = pocket_protector.__main__:main']},
    install_requires=['attrs',
                      'boltons',
                      'PyNaCl',
                      'ruamel.yaml',
                      'schema',
                      'face>=20.1.0']
)

"""
Release process:

* tox
* git commit (if applicable)
* Remove dev suffix from pocket_protector/_version.py version
* git commit -a -m "bump version for vX.Y.Z release"
* python setup.py sdist bdist_wheel upload
* git tag -a vX.Y.Z -m "brief summary"
* write CHANGELOG
* git commit
* bump pocket_protector/_version.py version onto n+1 dev
* git commit
* git push

Versions are of the format YY.MINOR.MICRO, see calver.org for more details.
"""



'''
Functions and classes for dealing with a checked-in file,
protected.yaml, which stores secret data securely.

There are two public classes: KeyFile, and Creds.
'''
import os
import re
import base64
import collections
import datetime
import hashlib
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

import attr
import nacl.utils
import nacl.public
import nacl.secret
import nacl.pwhash
import schema
import ruamel.yaml
from boltons.dictutils import OMD
from boltons.fileutils import atomic_save

try:
    unicode = unicode
except NameError:
    unicode = str  # py3


_VALID_NAME_RE = re.compile(r"^[A-z][-_A-z0-9]*\Z")


class PPError(Exception):
    pass


class PPKeyError(PPError, KeyError):
    pass


# _coerce to handle ruamel type switcheroo
def _coerce(type_, sub_schema):
    return schema.And(schema.Use(type_), sub_schema)


def _as_d(sub_schema): return _coerce(dict, sub_schema)
def _as_l(sub_schema): return _coerce(list, sub_schema)


_FILE_SCHEMA = schema.Schema(_as_d(
{
    "audit-log": _as_l([str]),
    "key-custodians": _as_d({
        schema.Optional(str): _as_d({
            "pwdkm": str,
        }),
    }),
    schema.Optional(schema.Regex("^(?!meta).*$")): _as_d({
        # allow string names for security domains,
        # but meta is reserved
        "meta": _as_d({
            "owners": _as_d({str: str}),
            "public-key": str,
        }),
        schema.Optional(schema.Regex("secret-.*")): str,
    }),
}))


OPSLIMIT = nacl.pwhash.argon2id.OPSLIMIT_SENSITIVE
MEMLIMIT = nacl.pwhash.argon2id.MEMLIMIT_MODERATE

# NOTE: this is a public class since it must be passed in
@attr.s(frozen=True)
class Creds(object):
    "Stores credentials used to open a KeyFile"
    name = attr.ib(validator=attr.validators.instance_of(unicode))
    passphrase = attr.ib(validator=attr.validators.instance_of(unicode))
    name_source = attr.ib(default=None)
    passphrase_source = attr.ib(default=None)


def _kdf(creds, salt):
    name = creds.name.encode('utf8')
    passphrase = creds.passphrase.encode('utf8')

    valet_key = hashlib.sha512(passphrase + salt + name).digest()
    # valet key can be used to share credentials
    # without exposing password
    return nacl.pwhash.argon2id.kdf(
        nacl.public.PrivateKey.SIZE,
        valet_key, hashlib.sha512(salt + name).digest()[:16],
        opslimit=OPSLIMIT,
        memlimit=MEMLIMIT)


def _decode(b64):
    '''
    assert everything is version 0
    later on for e.g. algorithm flexibility
    version 1, 2, 3 could be added with object
    specific encoder / decoder
    (this would be a lot of code work, but the files would
    be forwards compatible, and backwards can at least
    detect the problem and notify the user cleanly)
    '''
    raw = base64.b64decode(b64)
    if raw[:1] == b'\0':
        return raw[1:]
    raise PPError('version %s object not supported' % ord(raw[:1]))


def _encode(raw):
    'add version 0 byte to everything'
    return base64.b64encode(b'\0' + raw).decode('utf8')


@attr.s(frozen=True)
class _KeyCustodian(object):
    '''
    represents a key-custodian, who may be granted ownership
    (aka the ability to decrypt secrets) in one or more domains
    '''
    name = attr.ib()
    _public_key = attr.ib()
    _salt = attr.ib()

    def encrypt_for(self, bytes):
        'encrypt the passed bytes so that this key-custodian can decrypt'
        return nacl.public.SealedBox(self._public_key).encrypt(bytes)

    def decrypt_as(self, creds, bytes):
        'decrypt the passed bytes that were encrypted for this key-custodian'
        assert creds.name == self.name
        return nacl.public.SealedBox(
            nacl.public.PrivateKey(_kdf(creds, self._salt))).decrypt(bytes)

    @classmethod
    def from_creds(cls, creds):
        'create a new user based on new credentials'
        salt = os.urandom(8)
        private_key = nacl.public.PrivateKey(_kdf(creds, salt))
        return cls(
            name=creds.name, public_key=private_key.public_key, salt=salt)

    @classmethod
    def from_data(cls, name, data):
        # password derived key material
        pwdkm = _decode(data['pwdkm'])
        salt, public_key = pwdkm[:8], pwdkm[8:8 + nacl.public.PublicKey.SIZE]
        return cls(
            name=name, public_key=nacl.public.PublicKey(public_key), salt=salt)

    def as_data(self):
        return {
            'pwdkm': _encode(self._salt + self._public_key.encode()),
        }


@attr.s(frozen=True)
class _Owner(object):
    'represents an ownership relationship between a key-custodian and a key-domain'
    _name = attr.ib()
    _enc_domain_private_key = attr.ib()

    @classmethod
    def from_custodian_and_pkey(cls, key_custodian, pkey):
        '''
        create an ownership relationship based on a key_custodian
        and decrypted private key
        '''
        return cls(key_custodian.name, key_custodian.encrypt_for(pkey.encode()))

    def decrypt_private_key_bytes(self, creds, key_custodian):
        'decrypt the private key based on the passphrase'
        return key_custodian.decrypt_as(creds, self._enc_domain_private_key)

    @classmethod
    def from_data(cls, name, encrypted_private_key_bytes):
        return cls(name, _decode(encrypted_private_key_bytes))

    def as_data(self):
        return _encode(self._enc_domain_private_key)


def _err_map_attrib(item_name):
    'utility for giving good error messages'
    class MissingErrDict(dict):
        def __missing__(self, key):
            raise PPKeyError("no {0} of name {1} (known {0}s are {2})".format(
                item_name, key, ", ".join(self)))
    return attr.ib(default=attr.Factory(dict), converter=MissingErrDict)


def _deleted(mapping, key):
    '''
    like sort() vs sorted(), return a dict copy of mapping
    with key del'd out
    '''
    ret = dict(mapping)
    del ret[key]
    return ret


def _setitem(mapping, key, val):
    '''
    returns a dict-copy of mapping with key set to val
    '''
    ret = dict(mapping)
    ret[key] = val
    return ret


@attr.s(frozen=True)
class _EncryptedKeyDomain(object):
    'Represents a key domain with all values encrypted.'
    _name = attr.ib()
    _pub_key = attr.ib()
    _secrets = _err_map_attrib('secret')
    _owners = _err_map_attrib('owner')

    def _decrypt_private_key(self, key_custodian, creds):
        return nacl.public.PrivateKey(
            self._owners[creds.name].decrypt_private_key_bytes(
                creds, key_custodian))

    def get_decrypted(self, key_custodian, creds):
        if creds.name not in self._owners:
            raise PPError('{} is not an owner of {}'.format(
                creds.name, self._name))
        box = nacl.public.SealedBox(self._decrypt_private_key(
            key_custodian, creds))
        secrets = {}
        for name, val in self._secrets.items():
            secrets[name] = box.decrypt(val).decode('utf8')
        return _KeyDomain(secrets)

    def set_secret(self, name, value):
        'return a copy of the EncryptedKeyDomain with the new secret name/value'
        name_match = _VALID_NAME_RE.match(name)
        if not name_match:
            raise ValueError('valid secret names must begin with a letter, and'
                             ' consist only of ASCII letters, digits, and'
                             ' underscores, not: %r' % name)

        secrets = dict(self._secrets)
        box = nacl.public.SealedBox(self._pub_key)
        secrets[name] = box.encrypt(value.encode('utf8'))
        return attr.evolve(self, secrets=secrets)

    def add_secret(self, name, value):
        'like set_secret, but errors if secret exists'
        if name in self._secrets:
            raise PPError('secret {} already exists in {}'.format(
                name, self._name))
        return self.set_secret(name, value)

    def update_secret(self, name, value):
        'like set_secret, but errors if secret doesnt exist'
        chk = self._secrets[name]  # for error msg
        return self.set_secret(name, value)

    def rm_secret(self, name):
        return attr.evolve(self, secrets=_deleted(self._secrets, name))

    def add_owner(self, cur_creds, cur_key_custodian, new_key_custodian):
        'add a new owner based on a current owners credentials'
        domain_private_key = self._decrypt_private_key(
            cur_key_custodian, cur_creds)
        owners = dict(self._owners)
        owners[new_key_custodian.name] = _Owner.from_custodian_and_pkey(
            new_key_custodian, domain_private_key)
        return attr.evolve(self, owners=owners)

    def rm_owner(self, key_custodian_name):
        'remove owner, checking that domain has at least one user'
        if key_custodian_name not in self._owners:
            raise PPError("{} not an owner of {} (owners are {})".format(
                key_custodian_name, self._name, ", ".join(self._owners)))
        if len(self._owners) == 1:
            raise PPError(
                "cannot delete last owner {} from {} "
                "(secrets would be irretrievable)".format(key_custodian_name, self._name))
        return attr.evolve(self, owners=_deleted(self._owners, key_custodian_name))

    def get_owner_names(self):
        return list(self._owners)

    @classmethod
    def from_owner(cls, name, key_custodian):
        'create a new (empty) EncryptedKeyDomain with an initial owner'
        domain_private_key = nacl.public.PrivateKey.generate()
        return cls(
            name=name,
            pub_key=domain_private_key.public_key,
            secrets={},
            owners={
                key_custodian.name:
                    _Owner.from_custodian_and_pkey(
                        key_custodian, domain_private_key)})

    @classmethod
    def from_data(cls, name, data):
        'convert nested dict/list/str to instance'
        return cls(
            name=name,
            pub_key=nacl.public.PublicKey(
                _decode(data['meta']['public-key'])),
            owners={
                name: _Owner.from_data(name, owner_data)
                for name, owner_data in data['meta']['owners'].items()},
            secrets={
                name.split('secret-', 1)[1]: _decode(val)
                for name, val in data.items()
                if name.startswith('secret-')})

    def as_data(self):
        'convert instance to nested dict/list/str'
        data = { "secret-" + name: _encode(val)
                 for name, val in self._secrets.items() }
        # ensure keys go in sorted order
        data = collections.OrderedDict(sorted(data.items()))
        data['meta'] = collections.OrderedDict([
            ("public-key", _encode(self._pub_key.encode())),
            ("owners", collections.OrderedDict(
                sorted([(name, owner.as_data()) for name, owner in self._owners.items()]))),
        ])
        return data


class _KeyDomain(dict):
    'Represents a decrypted key domain which secrets can be read from'
    def __missing__(self, key):
        raise PPKeyError("no secret {} (known secrets are {})".format(
            key, ", ".join(self)))


def _represent_ordereddict(dumper, data):
    value = []
    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)
        value.append((node_key, node_value))
    return ruamel.yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)


ruamel.yaml.representer.RoundTripRepresenter.add_representer(
    collections.OrderedDict, _represent_ordereddict)


@attr.s(frozen=True)
class KeyFile(object):
    '''
    Represents a key-file (containing many domains)
    Can be read from and written to disk
    '''
    path = attr.ib()
    _domains = _err_map_attrib('domain')
    _key_custodians = _err_map_attrib('key custodian')
    _log = attr.ib(default=attr.Factory(list))
    _yaml = ruamel.yaml.YAML()  # class var
    _yaml.width = 100

    @classmethod
    def from_file(cls, path):
        'create a new KeyFile from path'
        with open(path, 'rb') as file:
            contents = file.read().decode('utf8')
        return cls.from_contents_and_path(contents, path)

    @classmethod
    def from_contents_and_path(cls, bytes, path):
        'create a new KeyFile from file contents'
        contents = cls._yaml.load(bytes)
        _FILE_SCHEMA.validate(contents)
        log = contents.pop('audit-log')
        key_custodians = {
            name: _KeyCustodian.from_data(name, val)
            for name, val in contents.pop('key-custodians').items()}
        encrypted_domains = {
            name: _EncryptedKeyDomain.from_data(name, data)
            for name, data in contents.items() }
        return cls(
            path=path, domains=encrypted_domains,
            key_custodians=key_custodians, log=log)

    def get_domain_names(self):
        return sorted(self._domains.keys())

    def get_domain_secret_names(self, domain_name):
        domain = self._domains[domain_name]
        return sorted(domain._secrets.keys())

    def get_all_secret_names(self):
        "return a map of secret names to names of domains that contain that secret"
        res = OMD()
        for domain_name, domain in self._domains.items():
            secrets_dict = domain._secrets
            for secret_name in secrets_dict:
                res.add(secret_name, domain_name)
        return dict(res)

    def get_contents(self):
        data = collections.OrderedDict(sorted([
            (domain_name, domain.as_data())
            for domain_name, domain in self._domains.items()]))
        data['key-custodians'] = collections.OrderedDict(sorted([
            (name, kc.as_data())
            for name, kc in self._key_custodians.items()]))
        data['audit-log'] = self._log
        stream = StringIO()
        self._yaml.dump(data, stream)
        text = stream.getvalue()
        return text

    def _new_log(self, entry, *a, **kw):
        cur_time_str = datetime.datetime.utcnow().isoformat().split('.')[0] + 'Z'
        new_entry = cur_time_str + ' -- ' + (entry.format(*a, **kw))
        return self._log + [new_entry]

    def get_audit_log(self):
        return list(self._log)

    def write(self):  # TODO: need way to get contents.
        'write contents to file'
        contents = self.get_contents()
        with atomic_save(self.path) as file:
            file.write(contents.encode('utf8'))
        return

    def add_domain(self, domain_name, key_custodian_name):
        '''
        return a copy with a new domain, empty but with one initial key custodian
        owner who can add other owners
        '''
        if domain_name in self._domains:
            raise PPError('tried to add domain that already exists: {}'.format(domain_name))
        key_custodian = self._key_custodians[key_custodian_name]
        domains = dict(self._domains)
        domains[domain_name] = _EncryptedKeyDomain.from_owner(
            domain_name, key_custodian)
        return attr.evolve(
            self, domains=domains,
            log=self._new_log('created domain {} with owner {}',
                              domain_name, key_custodian_name))

    def rm_domain(self, domain_name):
        '''
        return a copy with domain domain_name removed
        '''
        return attr.evolve(
            self, domains=_deleted(self._domains, domain_name),
            log=self._new_log('deleted domain {}', domain_name))

    def set_secret(self, domain_name, name, value):
        'return a copy of the KeyFile with the given secret name and value added to a domain'
        domains = dict(self._domains)
        domains[domain_name] = self._domains[domain_name].set_secret(name, value)
        return attr.evolve(
            self, domains=domains,
            log=self._new_log('set secret {} in {}', name, domain_name))

    def add_secret(self, domain_name, name, value):
        'add a secret that doesnt exist yet'
        domains = dict(self._domains)
        domains[domain_name] = self._domains[domain_name].add_secret(name, value)
        return attr.evolve(
            self, domains=domains,
            log=self._new_log('added secret {} in {}', name, domain_name))

    def update_secret(self, domain_name, name, value):
        'update the value of a secret that already exists'
        domains = dict(self._domains)
        domains[domain_name] = self._domains[domain_name].update_secret(name, value)
        return attr.evolve(
            self, domains=domains,
            log=self._new_log('updated secret {} in {}', name, domain_name))

    def rm_secret(self, domain_name, name):
        'return a copy with secret removed from domain'
        domains = dict(self._domains)
        domains[domain_name] = self._domains[domain_name].rm_secret(name)
        return attr.evolve(
            self, domains=domains,
            log=self._new_log('removed secret {} from {}', name, domain_name))

    def add_owner(self, domain_name, key_custodian_name, creds):
        '''
        Register a new key custodian owner of domain_name based on the
        credentials of an existing owner
        '''
        domains = dict(self._domains)
        domains[domain_name] = self._domains[domain_name].add_owner(
            cur_creds=creds, cur_key_custodian=self._key_custodians[creds.name],
            new_key_custodian=self._key_custodians[key_custodian_name])
        return attr.evolve(
            self, domains=domains,
            log=self._new_log('{} added owner {} to {}',
                creds.name, key_custodian_name, domain_name))

    def rm_owner(self, domain_name, key_custodian_name):
        '''
        Remove an owner from domain.
        (NOTE: due to file history, the removed owner
        will still be able to get to values until you rotate
        the domain keypair, and secret values)
        '''
        return attr.evolve(
            self, domains=_setitem(
                self._domains, domain_name,
                self._domains[domain_name].rm_owner(key_custodian_name)),
            log=self._new_log('removed owner {} from {}',
                key_custodian_name, domain_name))

    def add_key_custodian(self, creds):
        key_custodians = dict(self._key_custodians)
        if creds.name in key_custodians:
            raise PPError(
                'tried to add key custodian that already exists: {}'.format(creds.name))
        key_custodians[creds.name] = _KeyCustodian.from_creds(creds)
        return attr.evolve(
            self, key_custodians=key_custodians,
            log=self._new_log('created key custodian {}', creds.name))

    def rm_key_custodian(self, key_custodian_name):
        'remove key custodian and all domain ownerships'
        key_custodians = dict(self._key_custodians)
        domains = dict(self._domains)
        owned = []
        for name, domain in self._domains.items():
            if key_custodian_name in domain.get_owner_names():
                domains[name] = domain.rm_owner(key_custodian_name)
                owned.append(name)
        del key_custodians[key_custodian_name]
        return attr.evolve(
            self, key_custodians=key_custodians, domains=domains,
            log=self._new_log('removed key custodian {} (was owner of {})',
                key_custodian_name, ", ".join(owned)))

    def decrypt_domain(self, domain_name, creds):
        return self._domains[domain_name].get_decrypted(
            self._key_custodians[creds.name], creds)

    def set_key_custodian_passphrase(self, creds, new_passphrase):
        new_kc = _KeyCustodian.from_creds(Creds(creds.name, new_passphrase))
        cur_kc = self._key_custodians[creds.name]
        key_custodians = dict(self._key_custodians)
        key_custodians[creds.name] = new_kc
        domains = dict(self._domains)
        updated = []
        for name, domain in domains.items():
            if creds.name in domain.get_owner_names():
                domains[name] = self._domains[name].add_owner(
                    cur_creds=creds, cur_key_custodian=cur_kc,
                    new_key_custodian=new_kc)
                updated.append(name)
        return attr.evolve(
            self, key_custodians=key_custodians, domains=domains,
            log=self._new_log(
                'updated key custodian passphrase for {} (updated domains -> {})',
                    creds.name, ", ".join(updated)))

    def check_creds(self, creds):
        try:
            key_custodian = self._key_custodians[creds.name]
        except KeyError:
            return False
        try:
            key_custodian.decrypt_as(creds, key_custodian.encrypt_for(b'\0'))
        except Exception:  # TODO: what crypto error?
            return False
        return True

    def rotate_domain_key(self, domain_name, creds):
        '''
        rotate the keypair used to secure a domain
        NIST recommends keys be rotated and not kept in use for more than ~1-3 years
        see http://nvlpubs.nist.gov/nistpubs/SpecialPublications/NIST.SP.800-57pt1r4.pdf
        Recommendation for Key Management, Part 1: General
        section 5.3.6 Cryptoperiod Recommendations for Specific Key Types
        '''
        cur_domain = self._domains[domain_name]
        key_custodian = self._key_custodians[creds.name]
        cur_secrets = cur_domain.get_decrypted(key_custodian, creds)
        new_domain = _EncryptedKeyDomain.from_owner(domain_name, key_custodian)
        for name, val in cur_secrets.items():
            new_domain = new_domain.set_secret(name, val)
        for owner_name in cur_domain.get_owner_names():
            new_domain = new_domain.add_owner(
                cur_creds=creds, cur_key_custodian=key_custodian,
                new_key_custodian=self._key_custodians[owner_name])
        domains = dict(self._domains)
        domains[domain_name] = new_domain
        return attr.evolve(
            self, domains=domains,
            log=self._new_log('rotated key for domain {}', domain_name))

    def truncate_audit_log(self, max_keep):
        max_keep = int(max_keep)
        if len(self._log) < max_keep:
            return self
        msg = 'truncated %s audit log entries' % (len(self._log) - max_keep)
        new_log = [msg] + self._log[-max_keep:]
        return attr.evolve(self, log=new_log)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import pytest

from pocket_protector import file_keys

import tempfile


def test_file_keys(_fast_crypto):
    bob_creds = file_keys.Creds('bob@example.com', 'super-secret')
    alice_creds = file_keys.Creds('alice@example.com', 'super-duper-secret')

    _prev = [None]
    def chk(fk):
        assert fk.from_contents_and_path(fk.get_contents(), fk.path) == fk
        assert _prev[0] != fk, "function call resulted in no changes to data"
        _prev[0] = fk

    tmp = tempfile.NamedTemporaryFile()
    test1 = test = file_keys.KeyFile(path=tmp.name)
    chk(test)
    test2 = test = test.add_key_custodian(bob_creds)
    chk(test)
    test3 = test = test.add_domain('new_domain', bob_creds.name)
    chk(test)

    with pytest.raises(ValueError):
        test3f = test = test.add_secret('new_domain', '$brokenkey', 'world')

    test3a = test = test.add_secret('new_domain', 'hello', 'world')
    chk(test)
    test3b = test = test.update_secret('new_domain', 'hello', 'world2')
    chk(test)
    test4 = test = test.set_secret('new_domain', 'hello', 'world')
    chk(test)
    test4a = test = test.rm_secret('new_domain', 'hello')
    chk(test)
    test4b = test = test.set_secret('new_domain', 'hello', 'world')
    chk(test)
    assert test.decrypt_domain('new_domain', bob_creds)['hello'] == 'world'
    test5 = test = test.set_secret('new_domain', 'hello', 'better-world')
    chk(test)
    assert test.decrypt_domain('new_domain', bob_creds)['hello'] == 'better-world'
    test6 = test = test.add_key_custodian(alice_creds)
    chk(test)
    test7 = test = test.add_owner('new_domain', alice_creds.name, bob_creds)
    chk(test)
    test8 = _test = test.rm_owner('new_domain', alice_creds.name)
    chk(_test)  # throw away this mutation
    test9 = _test = test.rm_key_custodian(alice_creds.name)
    chk(_test)  # throw away this mutation
    test9a = _test = test.rm_domain('new_domain')
    chk(_test)
    before_rotate = test.decrypt_domain('new_domain', bob_creds)
    test10 = test = test.rotate_domain_key('new_domain', bob_creds)
    chk(test)
    assert test.decrypt_domain('new_domain', bob_creds) == before_rotate
    test11 = test = test.set_key_custodian_passphrase(bob_creds, 'ultra-extra-secret')
    test.write()
    round_trip = file_keys.KeyFile.from_file(test.path)
    assert round_trip == test
    print("generated file:")
    print(open(test.path).read())
    print("...")
